#######################################
# Example: ftDigiCam.py
#          Digital Camera with live
#          video stream to TXT display
#          and autofocus functionality
#          (c) 2016 by Torsten Stuehn
#          version 0.83 from 2016-03-13
#######################################

# Python2/3 'print' compatibility
from __future__ import print_function

import ftrobopy
import ftrobopytools
import os, sys
import time
import math
import threading
import threading
import SimpleHTTPServer
import SocketServer

# define simple webserver and provide direct access
# to pictures taken at: http://192.168.7.2:8085
class WebServer(threading.Thread):
  def __init__(self, document_root):
    self._document_root = document_root
    self._httpd_server = SocketServer.TCPServer(("", 8085), SimpleHTTPServer.SimpleHTTPRequestHandler)
    threading.Thread.__init__(self)
    return
  def __del__(self):
    self._httpd_server.server_close()
  def run(self):
    while True:
      currentDir=os.getcwd()
      os.chdir(self._document_root)
      self._httpd_server.handle_request()
      os.chdir(currentDir)
      time.sleep(0.5)

startWebServer = True

# initialize socket connection to TXT-Main
txt = ftrobopy.ftrobopy('127.0.0.1', 65000)

# define the speed of the focus motor
# (max:512, min:100, the slower the speed the better the focus)
focus_speed        = 512

# define the directory where the pictures should be stored
# if you want to store pictures at external SD card, set this to e.g. '/mnt'
picture_dir        = '/opt/knobloch/ftrobopy/PICTURES'
# override directory if it was given as command line argument
if len(sys.argv) > 1:
  picture_dir = sys.argv[1]

# define the prefix of the picture-filename
fname_prefix       = 'PIC'

# if set to 0, no live video is shown on TXT Display
displayLiveStream  = 1 # live video on TXT display, 0=no 1=yes

# define minimum and maximum size of list for contrast search
hist_minlength     = 10
hist_maxlength     = 30

# definition of TXT outputs and inputs
FocusMotor       = txt.motor(1) # the focus motor is connected to M1
Trigger          = txt.input(1) # the camera trigger with auto focus is connected to I1

if startWebServer:
  httpd_thread = WebServer(picture_dir)
  httpd_thread.setDaemon(True)
  httpd_thread.start()

if displayLiveStream:
  # make backup copy of TXT display contents
  with open('/dev/fb0', 'rb') as f:
    framebuffer_backup=f.read()

try:

  # initialize camera (/dev/video0)
  fps    = 15  # frames per second
  width  = 320 # width of camera image
  height = 240 # height of camera image
  videv  = ftrobopytools.camInit(fps, width, height, 0, 0)

  if displayLiveStream:
    # initialize Standard Display Library (SDL)
    ftrobopytools.sdlInit()
    # reset text/cmd console (compromized by SDL)
    os.system('reset')

  contrast                = 0
  hist_contrast           = [0]
  hist_counter            = [0]
  ave_contrast            = 0
  counter_hist_shift      = 2

  # choose rectangular region for contrast measurement
  xtopleft     =  120
  ytopleft     =  90
  xbottomright =  180
  ybottomright =  150
  
  state     = 0
  direction = 1
  for i in range(1000):
    contrast = ftrobopytools.measureContrast(videv,
                                             width, height,
                                             xtopleft, ytopleft, xbottomright, ybottomright,
                                             displayLiveStream)
    motor_counter = FocusMotor.getCurrentDistance()

    # if camera picture was invalid use last values
    if contrast == None:
      contrast = hist_contrast[-1]

    contrast = int(math.sqrt(contrast)*2+0.5)

    hist_contrast.append(contrast)

    # print state and contrast list for debugging
    # uncomment the next line to see what is going on
    #print("state:",state,hist_contrast)

    if len(hist_contrast) > hist_maxlength:
      hist_contrast = hist_contrast[1:]
    
    hist_counter.append(motor_counter)
    if len(hist_counter) > hist_maxlength:
      hist_counter = hist_counter[1:]

    max_contrast       = max(hist_contrast)
    min_contrast       = min(hist_contrast)
    if min_contrast == max_contrast:
      contrast_variation = 0
    else:
      contrast_variation = 1

    if state == 0:
      # savety fallback
      if hist_counter[-1] != hist_counter[-2]:
        FocusMotor.stop()
        time.sleep(0.02)
      if Trigger.state() != 0:
        # start increasing focus                                                                                                                                                    
        FocusMotor.setDistance(3000)                                                                                                                                                
        FocusMotor.setSpeed(focus_speed*direction)
        time.sleep(0.02)
        hist_contrast = [0]                                                                                                                                                         
        hist_counter  = [0]                                                                                                                                                         
        state  = 1

    if state == 1:
      if len(hist_contrast) > hist_minlength and ( (hist_contrast[-1] <= hist_contrast[-2]) and (hist_contrast[-2] < hist_contrast[-3]) or contrast_variation == 0 ):
        hist_contrast = [0]
        hist_counter  = [0]
        FocusMotor.stop()
        time.sleep(0.02)
        if contrast_variation == 0:
          state     = 0
          direction = -direction
        else:
          # start decreasing focus
          FocusMotor.setDistance(3000)
          FocusMotor.setSpeed(-focus_speed*direction)
          time.sleep(0.01)
          state = 2

    if state == 2:
      if len(hist_contrast) > hist_minlength and ( (hist_contrast[-1] <= hist_contrast[-2]) and (hist_contrast[-2] < hist_contrast[-3]) or contrast_variation == 0 ):
        FocusMotor.stop()
        time.sleep(0.02)
        if contrast_variation == 0:
          state     = 0
          direction = -direction
        else:
          # start increasing focus to maximum contrast
          idx = hist_contrast.index(max(hist_contrast))
          # counter_hist_shift takes care of lag shift between camera pics and motor counter
          bestfocus_counter = hist_counter[idx-counter_hist_shift]
          FocusMotor.setDistance(abs(hist_counter[-1]-bestfocus_counter))
          FocusMotor.setSpeed(focus_speed*direction)
          time.sleep(0.02)
          state = 3

    if state == 3:
      if FocusMotor.finished():
        state = 4

    if state == 4:
        # save jpeg in high resolution (1280x720)
        # close (low resolution) camera device
        ftrobopytools.camClose(videv, 0)
        # open (high resolution) camera device
        high_fps     = 5    # 5 is the lowest possible framerate of the TXT camera
        high_width   = 1280 # 1280 is the maximum horizontal resolution of the TXT camera
        high_height  = 720  # 720 is the maximum vertical resolution of the TXT camera
        videv = ftrobopytools.camInit(high_fps, high_width, high_height, 0, 0)
        # get high resolution snapshot as jpg image
        jpg   = ftrobopytools.getJPEGImage(videv)
        # close (high resolution) camera device
        ftrobopytools.camClose(videv, 0)
        # restore resolution for liveStreaming
        videv = ftrobopytools.camInit(fps, width, height, 0, 0)
        # save jpeg to file and increment picture count index
        try:
          with open(picture_dir+'/'+fname_prefix+'IDX','r') as f:
            pict_number = int(f.read())
        except:
          pict_number = 0
        with open(picture_dir+'/'+fname_prefix + '%04i' % pict_number +'.JPG', 'wb') as f:
          f.write(jpg)
        with open(picture_dir+'/'+fname_prefix+'IDX','w') as f:
          f.write(str(pict_number + 1))
        # ready for the next picture
        hist_contrast = [0]
        hist_counter  = [0]
        state = 0

except ftrobopytools as error:
  print(error)

finally:
  # close camera device
  ftrobopytools.camClose(videv, 0)
  if displayLiveStream:
    # close Standard Display Library
    ftrobopytools.sdlClose()
    # restore TXT display
    with open('/dev/fb0', 'wb') as f:
      f.write(framebuffer_backup)

